import { LexDefinition, LexRule } from "../def/language";
import { TokenizedInput } from "../def/token";
import { CallbackController } from "../parser/callback";
/**
 * 字句解析器用のinterface
 *
 * TODO: 要改善
 */
export interface ILexer {
    exec(str: string): Array<TokenizedInput>;
    setCallbackController(cc: CallbackController): void;
}
/**
 * 字句解析器
 * 入力を受け取ってトークン化する
 */
export declare class Lexer implements ILexer {
    private _rule_id;
    private _def;
    private _reset;
    private _input;
    private _next_index;
    /**
     * 次に読み込む入力の位置
     */
    readonly next_index: number;
    private _status;
    /**
     * 字句解析器の内部状態を取得する
     * @returns {"uninitialized" | "ready" | "finished"}
     */
    readonly status: "uninitialized" | "ready" | "finished";
    private callback_controller;
    /**
     * @param {LexDefinition} def 解析する入力の字句規則
     * @param {string} input 解析したい入力
     */
    constructor(def: LexDefinition, input?: string);
    /**
     * 入力を解析する際に呼び出すコールバックコントローラーを設定する
     * @param {CallbackController} cc 使用されるべきコールバックコントローラー
     */
    setCallbackController(cc: CallbackController): void;
    /**
     * 字句解析機のリセット時に現在の字句規則になるようにする
     */
    setCurrentDefinitionAsDefault(): void;
    /**
     * Lexerの内部状態をコンストラクタ呼び出し直後まで戻す
     */
    reset(input?: string): void;
    /**
     * 入力を与えて解析可能(step()を呼び出し可能)な状態にする
     * @param {string} input 解析する入力
     */
    init(input: string): void;
    /**
     * 入力からトークン1つ分読み込む
     * @returns {TokenizedInput} トークン化された入力
     */
    step(): TokenizedInput;
    /**
     * 与えられた入力をすべて解析し、トークン列を返す
     * @param {string} input 解析する入力
     * @returns {Array<TokenizedInput>} 入力から生成されたトークン列
     */
    exec(input?: string): Array<TokenizedInput>;
    /**
     * 字句規則を追加し、そのidを返す
     * @param {LexRule} rule 新しく追加するルール
     * @returns {number} 追加されたルールの持つid
     */
    add(rule: LexRule): number;
    /**
     * 字句規則を削除する
     * 削除に失敗した場合は例外をスローする
     *
     * TODO: もっとましな実装にする
     * @param {number} id 削除するルールのid
     * @returns {LexRule} 削除されたルール
     */
    del(id: number): LexRule;
}
