import { Token } from "../def/token";
import { GrammarDB } from "./grammardb";
/**
 * 単一のLRアイテムであり、`S -> A . B [$]` のようなアイテムの規則id・ドットの位置・先読み記号の集合の情報を持つ
 *
 * [[GrammarDB]]から与えられるトークンIDをもとにして、LR(0)およびLR(1)アイテムとしてのハッシュ値を生成することができる
 *
 * Immutableであるべきオブジェクトであるため、インスタンス生成後は内部状態が変化することはないと仮定される
 */
export declare class ClosureItem {
    private grammardb;
    private _rule_id;
    private _dot_index;
    private _lookaheads;
    private _lr0_hash;
    private _lr1_hash;
    /**
     * @param {GrammarDB} grammardb 使用する構文の情報
     * @param {number} _rule_id 構文のid
     * @param {number} _dot_index ドットの位置
     * @param {Array<Token>} _lookaheads 先読み記号の集合
     */
    constructor(grammardb: GrammarDB, _rule_id: number, _dot_index: number, _lookaheads: Array<Token>);
    /**
     * 自身の規則idを返す
     *
     * 規則idはルールの定義順に0,1,2,...と割り振られる
     * @returns {number}
     */
    readonly rule_id: number;
    /**
     * 現在の読み込み位置を意味するドットの位置を返す
     * @returns {number}
     */
    readonly dot_index: number;
    /**
     * LR(1)先読み記号の集合を配列として返す
     *
     * 配列のコピーではなく参照が返されるので、結果しとて得られた配列に変更を加えてはならない
     * @returns {Array<Token>}
     */
    readonly lookaheads: Array<Token>;
    /**
     * 先読み記号の配列を、[[GrammarDB]]によって割り振られるトークンid順にソートする
     */
    private sortLA();
    /**
     * ハッシュ文字列を生成する
     */
    private updateHash();
    /**
     * 先読み部分を除いたLR(0)アイテムとしてのハッシュ文字列を得る
     * @returns {string}
     */
    getLR0Hash(): string;
    /**
     * 先読み部分を含めたLR(1)アイテムとしてのハッシュ文字列を得る
     * @returns {string}
     */
    getLR1Hash(): string;
    /**
     * LR(0)ハッシュの一致を調べる
     * @param {ClosureItem} c 比較対象のLRアイテム
     * @returns {boolean}
     */
    isSameLR0(c: ClosureItem): boolean;
    /**
     * LR(1)ハッシュの一致を調べる
     * @param {ClosureItem} c 比較対象のLRアイテム
     * @returns {boolean}
     */
    isSameLR1(c: ClosureItem): boolean;
    /**
     * LR0部分を維持しながらLR1先読み記号ごとにClosureItemを分割し、先読み記号の数が1のClosureItemの集合を生成する
     */
    separateByLookAheads(): Array<ClosureItem>;
    /**
     * LR0部分が同じ2つのClosureItemについて、先読み部分を統合した新しいClosureItemを生成する
     *
     * 異なるLR(0)アイテムであった場合、nullを返す
     * @param {ClosureItem} c マージ対象のLRアイテム
     * @returns {ClosureItem | null} 先読み部分がマージされた新しいLRアイテム
     */
    merge(c: ClosureItem): ClosureItem | null;
}
