import { ClosureItem } from "./closureitem";
import { GrammarDB } from "./grammardb";
/**
 * 複数のLRアイテムを保持するアイテム集合であり、インスタンス生成時に自身をクロージャー展開する
 *
 * [[GrammarDB]]から与えられるトークンIDをもとにして、LR(0)およびLR(1)アイテム集合としてのハッシュ値を生成することができる
 *
 * Immutableであるべきオブジェクトであるため、インスタンス生成後は内部状態が変化することはないと仮定される
 */
export declare class ClosureSet {
    private grammardb;
    private closureset;
    private _lr0_hash;
    private _lr1_hash;
    /**
     * @param {GrammarDB} grammardb 使用する構文の情報
     * @param {Array<ClosureItem>} closureset
     */
    constructor(grammardb: GrammarDB, closureset: Array<ClosureItem>);
    /**
     * 自身が保持する複数の[[ClosureItem]]は、常にLR(1)ハッシュによってソートされた状態に保たれているようにする
     */
    private sort();
    /**
     * 保持しているLRアイテムの数
     */
    readonly size: number;
    /**
     * 保持している[[ClosureItem]]の配列を得る
     * @param {boolean} prevent_copy trueを与えると配列をコピーせず返す
     *
     * 得られた配列に変更が加えられないと保証される場合に用いる
     * @returns {Array<ClosureItem>}
     */
    getArray(prevent_copy?: boolean): Array<ClosureItem>;
    /**
     * LRアイテムが集合に含まれているかどうかを調べる
     *
     * @param {ClosureItem} item
     * @returns {boolean}
     */
    includes(item: ClosureItem): boolean;
    /**
     * LR(0)ハッシュの一致を調べる
     * @param {ClosureSet} cs 比較対象のアイテム集合
     * @returns {boolean}
     */
    isSameLR0(cs: ClosureSet): boolean;
    /**
     * LR(1)ハッシュの一致を調べる
     * @param {ClosureSet} cs 比較対象のアイテム集合
     * @returns {boolean}
     */
    isSameLR1(cs: ClosureSet): boolean;
    /**
     * ハッシュ文字列を生成する
     */
    private updateHash();
    /**
     * LR(0)アイテム集合としてのハッシュ文字列を得る
     * @returns {string}
     */
    getLR0Hash(): string;
    /**
     * LR(1)アイテム集合としてのハッシュ文字列を得る
     * @returns {string}
     */
    getLR1Hash(): string;
    /**
     * LR(0)部分が同じ2つのClosureSetについて、先読み部分を統合した新しいClosureSetを生成する
     *
     * 異なるLR(0)アイテム集合であった場合、nullを返す
     * @param {ClosureSet} cs マージ対象のアイテム集合
     * @returns {ClosureSet | null} 先読み部分がマージされた新しいアイテム集合
     */
    mergeLA(cs: ClosureSet): ClosureSet | null;
    /**
     * クロージャー展開を行う
     *
     * TODO: リファクタリング
     */
    private expandClosure();
}
