"use strict";
Object.defineProperty(exports, "__esModule", { value: true });
const token_1 = require("../def/token");
const closureitem_1 = require("./closureitem");
const closureset_1 = require("./closureset");
/**
 * 構文規則からLR(1)DFAおよびLALR(1)DFAを生成する
 */
class DFAGenerator {
    /**
     * @param {GrammarDB} grammardb 構文規則
     */
    constructor(grammardb) {
        this.grammardb = grammardb;
        this.generateDFA();
        this.mergeLA();
    }
    /**
     * LR(1)DFAを得る
     * @returns {DFA}
     */
    getLR1DFA() {
        return this.lr_dfa;
    }
    /**
     * LALR(1)DFAを得る
     * @returns {DFA}
     */
    getLALR1DFA() {
        return this.lalr_dfa;
    }
    /**
     * DFAの生成
     */
    generateDFA() {
        const initial_item = new closureitem_1.ClosureItem(this.grammardb, -1, 0, [token_1.SYMBOL_EOF]);
        const initial_set = new closureset_1.ClosureSet(this.grammardb, [initial_item]);
        const dfa = [{ closure: initial_set, edge: new Map() }];
        // 変更がなくなるまでループ
        let flg_changed = true;
        let i = 0;
        while (flg_changed) {
            flg_changed = false;
            while (i < dfa.length) {
                const closure = dfa[i].closure;
                const edge = dfa[i].edge;
                const new_sets = this.generateNewClosureSets(closure);
                // 与えられたDFANodeと全く同じDFANodeがある場合、そのindexを返す
                // 見つからなければ-1を返す
                const getIndexOfDuplicatedNode = (dfa, new_node) => {
                    let index = -1;
                    for (const [i, node] of dfa.entries()) {
                        if (new_node.closure.isSameLR1(node.closure)) {
                            index = i;
                            break;
                        }
                    }
                    return index;
                };
                // 新しいノードを生成する
                for (const [edge_label, cs] of new_sets) {
                    const new_node = { closure: cs, edge: new Map() };
                    // 既存のNodeのなかに同一のClosureSetを持つものがないか調べる
                    const duplicated_index = getIndexOfDuplicatedNode(dfa, new_node);
                    let index_to;
                    if (duplicated_index == -1) {
                        // 既存の状態と重複しない
                        dfa.push(new_node);
                        index_to = dfa.length - 1;
                        flg_changed = true;
                    }
                    else {
                        // 既存の状態と規則が重複する
                        // 新しいノードの追加は行わず、重複する既存ノードに対して辺を張る
                        index_to = duplicated_index;
                    }
                    // 辺を追加する
                    if (!edge.has(edge_label)) {
                        edge.set(edge_label, index_to);
                        // 新しい辺が追加された
                        flg_changed = true;
                        // DFAを更新
                        dfa[i] = { closure, edge };
                    }
                }
                i++;
            }
            i = 0;
        }
        this.lr_dfa = dfa;
    }
    /**
     * LR(1)オートマトンの先読み部分をマージして、LALR(1)オートマトンを作る
     */
    mergeLA() {
        if (this.lalr_dfa !== undefined || this.lr_dfa === undefined)
            return;
        const base = this.lr_dfa.slice(); // nullを許容する
        const merge_to = new Map(); // マージ先への対応関係を保持する
        for (let i = 0; i < base.length; i++) {
            if (base[i] === null)
                continue;
            for (let ii = i + 1; ii < base.length; ii++) {
                if (base[ii] === null)
                    continue;
                // LR(0)アイテムセット部分が重複
                if (base[i].closure.isSameLR0(base[ii].closure)) {
                    // ii番目の先読み部分をi番目にマージする
                    // インデックス番号の大きい方が削除される
                    // 辺情報は、削除された要素の持つ辺の対象もいずれマージされて消えるため操作しなくてよい
                    // 更新
                    // Nodeに変更をかけるとLR(1)DFAの中身まで変化してしまうため新しいオブジェクトを生成する
                    base[i] = { closure: base[i].closure.mergeLA(base[ii].closure), edge: base[i].edge };
                    // ii番目を削除
                    base[ii] = null;
                    // マージ元->マージ先への対応関係を保持
                    merge_to.set(ii, i);
                }
            }
        }
        // 削除した部分を配列から抜き取る
        const prev_length = base.length; // ノードをマージする前のノード総数
        const fix = new Array(prev_length); // (元のindex->現在のindex)の対応表を作る
        let d = 0; // ずれ
        // nullで埋めた部分を消すことによるindexの変化
        for (let i = 0; i < prev_length; i++) {
            if (base[i] === null)
                d += 1; // ノードが削除されていた場合、以降のインデックスを1つずらす
            else
                fix[i] = i - d;
        }
        // 配列からnull埋めした部分を削除したものを作る
        const shortened = [];
        for (const node of base) {
            if (node !== null)
                shortened.push(node);
        }
        // fixのうち、ノードが削除された部分を正しい対応で埋める
        for (const [from, to] of merge_to) {
            let index = to;
            while (merge_to.has(index))
                index = merge_to.get(index);
            if (index !== to)
                merge_to.set(to, index); // 対応表を更新しておく
            fix[from] = fix[index]; // toを繰り返し辿っているので未定義部分へのアクセスは発生しない
        }
        const result = [];
        // インデックスの対応表をもとに辺情報を書き換える
        for (const node of shortened) {
            const new_edge = new Map();
            for (const [token, node_index] of node.edge) {
                new_edge.set(token, fix[node_index]);
            }
            result.push({ closure: node.closure, edge: new_edge });
        }
        this.lalr_dfa = result;
    }
    /**
     * 既存のClosureSetから新しい規則を生成し、対応する記号ごとにまとめる
     * @param closureset
     */
    generateNewClosureSets(closureset) {
        const tmp = new Map();
        // 規則から新しい規則を生成し、対応する記号ごとにまとめる
        for (const { rule_id, dot_index, lookaheads } of closureset.getArray()) {
            const pattern = this.grammardb.getRuleById(rule_id).pattern;
            if (dot_index == pattern.length)
                continue; // .が末尾にある場合はスキップ
            const new_ci = new closureitem_1.ClosureItem(this.grammardb, rule_id, dot_index + 1, lookaheads);
            const edge_label = pattern[dot_index];
            let items;
            if (tmp.has(edge_label)) {
                // 既に同じ記号が登録されている
                items = tmp.get(edge_label);
            }
            else {
                // 同じ記号が登録されていない
                items = [];
            }
            items.push(new_ci);
            tmp.set(edge_label, items);
        }
        // ClosureItemの配列からClosureSetに変換
        const result = new Map();
        for (const [edge_label, items] of tmp) {
            result.set(edge_label, new closureset_1.ClosureSet(this.grammardb, items));
        }
        return result;
    }
}
exports.DFAGenerator = DFAGenerator;
